﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Markup;
using System.Xml;
using System.Xml.Linq;
using Microsoft.Win32;
using System.Globalization;
using System.Windows.Media;

namespace VectorGrafik
{
    public class DesignerCanvas : Canvas
    {
        public Style StyleForObjects;
        protected override Size MeasureOverride(Size constraint)
        {
            Size size = new Size();
            foreach (UIElement element in Children)
            {
                double left = Canvas.GetLeft(element);
                double top = Canvas.GetTop(element);
                left = double.IsNaN(left) ? 0 : left;
                top = double.IsNaN(top) ? 0 : top;

                element.Measure(constraint);

                Size desiredSize = element.DesiredSize;
                if (!double.IsNaN(desiredSize.Width) && !double.IsNaN(desiredSize.Height))
                {
                    size.Width = Math.Max(size.Width, left + desiredSize.Width);
                    size.Height = Math.Max(size.Height, top + desiredSize.Height);
                }
            }

            // add some extra margin
            size.Width += 10;
            size.Height += 10;
            return size;
        }
        internal IEnumerable<DesignerItem> SelectedItems
        {
            get
            {
                var selectedItems = from item in this.Children.OfType<DesignerItem>()
                                    where item.IsSelected == true
                                    select item;

                return selectedItems;
            }
        }
        internal void DeselectAll()
        {
            foreach (DesignerItem item in this.SelectedItems)
            {
                item.IsSelected = false;
            }
        }
        internal void BringToFront(DesignerItem Item)
        {
            // für alle Objekte eine Neusortierung vornehmen
            List<UIElement> childrenSorted = (from UIElement item in this.Children
                                              orderby Canvas.GetZIndex(item as UIElement) ascending
                                              select item as UIElement).ToList();

            List<UIElement> selectionSorted = (from item in SelectedItems
                                               orderby Canvas.GetZIndex(item as UIElement) ascending
                                               select item as UIElement).ToList();
            int i = 0;
            int j = 0;

            foreach (UIElement item in childrenSorted)
            {
                if (selectionSorted.Contains(item))
                {
                    int idx = Canvas.GetZIndex(item);
                    Canvas.SetZIndex(item, childrenSorted.Count - selectionSorted.Count + j++);
                }
                else
                {
                    Canvas.SetZIndex(item, i++);
                }
            }            
            this.UpdateLayout();
        }
        internal void SendToBack(DesignerItem Item)
        {
            // für alle Objekte eine Neusortierung vornehmen
            List<UIElement> childrenSorted = (from UIElement item in this.Children
                                              orderby Canvas.GetZIndex(item as UIElement) ascending
                                              select item as UIElement).ToList();

            List<UIElement> selectionSorted = (from item in SelectedItems
                                               orderby Canvas.GetZIndex(item as UIElement) ascending
                                               select item as UIElement).ToList();
            int i = 0;
            int j = 0;
            foreach (UIElement item in childrenSorted)
            {
                if (selectionSorted.Contains(item))
                {
                    int idx = Canvas.GetZIndex(item);
                    Canvas.SetZIndex(item, j++);

                }
                else
                {
                    Canvas.SetZIndex(item, selectionSorted.Count + i++);
                }
            }
        }
        internal void BringForward(DesignerItem Item)
        {

            List<UIElement> ordered = (from item in SelectedItems
                                               orderby Canvas.GetZIndex(item as UIElement) ascending
                                               select item as UIElement).ToList();

            int count = this.Children.Count;

            for (int i = 0; i < ordered.Count; i++)
            {
                int currentIndex = Canvas.GetZIndex(ordered[i]);
                int newIndex = Math.Min(count - 1 - i, currentIndex + 1);
                if (currentIndex != newIndex)
                {
                    Canvas.SetZIndex(ordered[i], newIndex);
                    IEnumerable<UIElement> it = this.Children.OfType<UIElement>().Where(item => Canvas.GetZIndex(item) == newIndex);

                    foreach (UIElement elm in it)
                    {
                        if (elm != ordered[i])
                        {
                            Canvas.SetZIndex(elm, currentIndex);
                            break;
                        }
                    }
                }
            }
        }
        internal void SendBackward(DesignerItem Item)
        {
            List<UIElement> ordered = (from item in SelectedItems
                                       orderby Canvas.GetZIndex(item as UIElement) ascending
                                       select item as UIElement).ToList();

            int count = this.Children.Count;

            for (int i = 0; i < ordered.Count; i++)
            {
                int currentIndex = Canvas.GetZIndex(ordered[i]);
                int newIndex = Math.Max(i, currentIndex - 1);
                if (currentIndex != newIndex)
                {
                    Canvas.SetZIndex(ordered[i], newIndex);
                    IEnumerable<UIElement> it = this.Children.OfType<UIElement>().Where(item => Canvas.GetZIndex(item) == newIndex);

                    foreach (UIElement elm in it)
                    {
                        if (elm != ordered[i])
                        {
                            Canvas.SetZIndex(elm, currentIndex);
                            break;
                        }
                    }
                }
            }
        }
        protected override void OnDrop(DragEventArgs e)
        {
            base.OnDrop(e);
            string xamlString = e.Data.GetData("DESIGNER_ITEM") as string;
            if (!String.IsNullOrEmpty(xamlString))
            {
                DesignerItem newItem = null;
                FrameworkElement content = XamlReader.Load(XmlReader.Create(new StringReader(xamlString))) as FrameworkElement;

                if (content != null)
                {
                    newItem = new DesignerItem(this);
                    newItem.Content = content;

                    Point position = e.GetPosition(this);
                    if (content.MinHeight != 0 && content.MinWidth != 0)
                    {
                        newItem.Width = content.MinWidth * 2; ;
                        newItem.Height = content.MinHeight * 2;
                    }
                    else
                    {
                        newItem.Width = 65;
                        newItem.Height = 65;
                    }
                    DesignerCanvas.SetLeft(newItem, Math.Max(0, position.X - newItem.Width / 2));
                    DesignerCanvas.SetTop(newItem, Math.Max(0, position.Y - newItem.Height / 2));
                    this.Children.Add(newItem);

                    this.DeselectAll();
                    newItem.Style = Style;
                    newItem.IsSelected = true;
                }

                e.Handled = true;
            }
        }
        internal void Save_Executed(string FileName)
        {
            IEnumerable<DesignerItem> designerItems = this.Children.OfType<DesignerItem>();
            XElement designerItemsXML = SerializeDesignerItems(designerItems);
            XElement root = new XElement("Root");
            root.Add(designerItemsXML);
            SaveFile(root, FileName);
        }
        internal string SaveAs_Executed()
        {
            IEnumerable<DesignerItem> designerItems = this.Children.OfType<DesignerItem>();
            XElement designerItemsXML = SerializeDesignerItems(designerItems);
            XElement root = new XElement("Root");
            root.Add(designerItemsXML);
            return SaveFileAs(root);
        }
        internal string Open_Executed()
        {
            string fileName = "";
            XElement root = LoadSerializedDataFromFile(out fileName);
            if (root != null)
            {

                this.Children.Clear();

                IEnumerable<XElement> itemsXML = root.Elements("DesignerItems").Elements("DesignerItem");
                foreach (XElement itemXML in itemsXML)
                {
                    DesignerItem item = DeserializeDesignerItem(itemXML, 0, 0);
                    this.Children.Add(item);
                }

                this.InvalidateVisual();
            }
            return fileName;
        }
        private XElement SerializeDesignerItems(IEnumerable<DesignerItem> designerItems)
        {
            XElement serializedItems = new XElement("DesignerItems",
                                       from item in designerItems
                                       let contentXaml = XamlWriter.Save(((DesignerItem)item).Content)
                                       let contentTransform=XamlWriter.Save(((DesignerItem) item).RenderTransform)
                                       select new XElement("DesignerItem",
                                                  new XElement("Left", Canvas.GetLeft(item)),
                                                  new XElement("Top", Canvas.GetTop(item)),
                                                  new XElement("Width", item.Width),
                                                  new XElement("Height", item.Height),
                                                  new XElement("zIndex", Canvas.GetZIndex(item)),
                                                  new XElement("Content", contentXaml),
                                                  new XElement ("Transform",contentTransform)
                                              )
                                   );

            return serializedItems;
        }
        private DesignerItem DeserializeDesignerItem(XElement itemXML, double OffsetX, double OffsetY)
        {
            DesignerItem item = new DesignerItem(this);
            item.Width = Double.Parse(itemXML.Element("Width").Value, CultureInfo.InvariantCulture);
            item.Height = Double.Parse(itemXML.Element("Height").Value, CultureInfo.InvariantCulture);
            Canvas.SetLeft(item, Double.Parse(itemXML.Element("Left").Value, CultureInfo.InvariantCulture) + OffsetX);
            Canvas.SetTop(item, Double.Parse(itemXML.Element("Top").Value, CultureInfo.InvariantCulture) + OffsetY);
            Canvas.SetZIndex(item, Int32.Parse(itemXML.Element("zIndex").Value));
            Object content = XamlReader.Load(XmlReader.Create(new StringReader(itemXML.Element("Content").Value)));
            Object transform = XamlReader.Load(XmlReader.Create(new StringReader(itemXML.Element("Transform").Value)));
            item.Content = content;
            item.RenderTransform = (Transform) transform;
            item.Style = Style;
            return item;
        }
        private void SaveFile(XElement xElement, string FileName)
        {
                try
                {
                    xElement.Save(FileName);
                }
                catch (Exception ex)
                {
                    MessageBox.Show(ex.StackTrace, ex.Message, MessageBoxButton.OK, MessageBoxImage.Error);
                }
        }
        private string SaveFileAs(XElement xElement)
        {
            string fileName = "";
            SaveFileDialog saveFile = new SaveFileDialog();
            saveFile.Filter = "Grafiker (*.grf)|*.grf";
            saveFile.DefaultExt = "grf";
            if (saveFile.ShowDialog() == true)
            {
                try
                {
                    xElement.Save(saveFile.FileName);
                    fileName = saveFile.FileName;
                }
                catch (Exception ex)
                {
                    MessageBox.Show(ex.StackTrace, ex.Message, MessageBoxButton.OK, MessageBoxImage.Error);
                }
            }
            return fileName;
        }
        private XElement LoadSerializedDataFromFile(out string FileName)
        {
            FileName = "";
            OpenFileDialog openFile = new OpenFileDialog();
            openFile.Filter = "Grafiker (*.grf)|*.grf";
            openFile.DefaultExt = "grf";
            if (openFile.ShowDialog() == true)
            {
                try
                {
                    FileName = openFile.FileName;
                    return XElement.Load(openFile.FileName);
                }
                catch (Exception e)
                {
                    MessageBox.Show(e.StackTrace, e.Message, MessageBoxButton.OK, MessageBoxImage.Error);
                }
            }

            return null;
        }
        internal void Delete(DesignerItem designerItem)
        {
            this.Children.Remove(designerItem);    
        }
    }
}
